



deps --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
deps --version
IF %ERRORLEVEL% NEQ 0 exit /B 1
python -m deps --version
IF %ERRORLEVEL% NEQ 0 exit /B 1
pytest source/python
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
